#define _WIN32_WINNT 0x0400 // for mouse wheel

#include <windows.h>
#include <windowsx.h>

#include "D3DApp.h"
#include "udebug.h"
#include "UUtil.h"

//
// Global access to the app (needed for the global WndProc())
//
static D3DApp* g_pd3dApp = NULL;

char strFrameRate[32] = "";

//
// constructor
//
D3DApp::D3DApp()
{
    g_pd3dApp = this;

    m_pd3d = NULL;
    m_pd3dDevice = NULL;
}

//
// clean up resources
//
void D3DApp::cleanUp()
{
    m_map.release();
    m_char.release();

    releaseLog();

    ReleaseCom( m_pd3dDevice );
    ReleaseCom( m_pd3d );
}

LRESULT WINAPI WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    return g_pd3dApp->msgProc( hWnd, msg, wParam, lParam );
}

//
// create a device and a window
//
HRESULT D3DApp::create()
{
    // register window's class
    WNDCLASSEX wc = {
        sizeof( WNDCLASSEX ), // cbSize
        CS_CLASSDC,           // style
        WndProc,              // lpfnWndProc
        0L,                   // cbClassExtra
        0L,                   // cbWndExtra
        GetModuleHandle(NULL),// hInstance
        NULL,                 // hIcon
        NULL,                 // hCursor
        NULL,                 // hbrBackground
        NULL,                 // lpszMenuName
        "SkinnedMesh",       // lpszClassName
        NULL                  // hIconSm
    };
    RegisterClassEx( &wc );
    
    // create the application's window
    HWND hWnd = CreateWindow( "SkinnedMesh", // lpClassName
                              "Skinned Mesh", // lpWindowName
                              WS_OVERLAPPED | WS_SYSMENU, // dwStyle
                              100, // x
                              100, // y
                              800, // nWidth
                              600, // nHeight
                              GetDesktopWindow(), // hWndParent
                              NULL, // hMenu
                              wc.hInstance, // hInstance
                              NULL ); // lpParam

    // Show the window
    ShowWindow( hWnd, SW_SHOWDEFAULT );
    UpdateWindow( hWnd );

    // create the D3D object
    if( NULL == (m_pd3d = Direct3DCreate9( D3D_SDK_VERSION )) )
    {
        return E_FAIL;
    }

    // set up the structure used to create the D3DDevice
    D3DPRESENT_PARAMETERS m_d3dpp;
    ZeroMemory( &m_d3dpp, sizeof( m_d3dpp ) );
    m_d3dpp.Windowed = true;
    m_d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    m_d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

    m_d3dpp.BackBufferCount = 1;
    m_d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    m_d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    m_d3dpp.EnableAutoDepthStencil = TRUE;
    m_d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    m_d3dpp.BackBufferWidth = 800;
    m_d3dpp.BackBufferHeight = 600;

    // Create the D3DDevice
    if( FAILED( m_pd3d->CreateDevice( D3DADAPTER_DEFAULT,
                                      D3DDEVTYPE_HAL,
                                      hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &m_d3dpp,
                                      &m_pd3dDevice ) ) )
    {
        m_pd3d->Release();
        return E_FAIL;
    }

    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, D3DZB_TRUE );

    RECT logRegion;
    GetClientRect( hWnd, &logRegion );
    // initialize log modules
    initLog( m_pd3dDevice, logRegion, 20 );

    // initialize timer
    m_timer.init();

    // initialize a map
    // get a name of a map file
    char fname[1024];
    memset( fname, 0, sizeof(char)*1024 );

    OPENFILENAME ofn;
    memset(&ofn, 0, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hWnd;
    ofn.lpstrFilter = "map file (*.map)\0*.map\0BMP map file (*.bmp)\0*.bmp\0All\0*.*\0";
    ofn.lpstrDefExt = "map";
    ofn.lpstrFile = fname;
    ofn.nMaxFile = 255;
    ofn.lpstrFileTitle = "Open Map File";
    ofn.Flags = OFN_OVERWRITEPROMPT;

    if (GetOpenFileName(&ofn)==TRUE)
    {
        m_map.load( m_pd3dDevice, fname );
    } else {
        m_map.create( m_pd3dDevice, 15, 15 );
    }

    // initialize a character
    if( FAILED(m_char.initDeviceObjects( m_pd3dDevice, "tiny_4anim.x", &m_map )) )
    {
        return E_FAIL;
    }

    // initialize a camera
    m_camera.init( m_char.getPosition() );

    // setup lights
    setupLights();

    return S_OK;
}

//
// run a message loop
//
INT D3DApp::run()
{
    MSG msg;
    ZeroMemory( &msg, sizeof(msg) );
    while( msg.message != WM_QUIT )
    {
        if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
        {
            TranslateMessage( &msg );
            DispatchMessage( &msg );
        }
        else
        {
            float elapsedTime = m_timer.getElapsedTime();

            frameMove( elapsedTime );
            render();

        }
    }

    return (INT)msg.wParam;
}

//
// message process
// 
LRESULT D3DApp::msgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    static int oldx = 0, oldy = 0;
    int x, y;
    switch( msg )
    {
    case WM_DESTROY :
        cleanUp();
        PostQuitMessage( 0 );
        return 0;

    case WM_LBUTTONDOWN :
    case WM_RBUTTONDOWN :
    case WM_MOUSEMOVE :
        // get the position of a cursor
        x = GET_X_LPARAM( lParam );
        y = GET_Y_LPARAM( lParam );
        // rotate a camera
        if( wParam == MK_RBUTTON )
        {
            // calculate the difference between new and old position of mouse
            if( oldx != 0 && oldy != 0 )
            {
                // rotate camera
                RECT cRect;
                GetClientRect( hWnd, &cRect );
                m_camera.rotate( (float)(oldx - x)/(cRect.right-cRect.left)*4, (float)(y - oldy)/(cRect.bottom-cRect.top)*4 );
            }
            // store old position of mouse
            oldx = x;
            oldy = y;
        }
        // move a character
        else if( wParam == MK_LBUTTON )
        {
            // find a picked position
            D3DXVECTOR3 vPicked;
            if( findPickedPos( x, y, &vPicked ) )
            {
                m_char.move( vPicked );
            }
        }
        break;
    case WM_RBUTTONUP :
        oldx = 0;
        oldy = 0;
        break;
    case WM_MOUSEWHEEL :
        {
            // get wheel delta
            short zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
            float delta = (float)(zDelta / WHEEL_DELTA);
            // zoom a camera
            m_camera.zoom( delta );
        }
        break;
#ifdef PROF_RENDER
    case WM_CHAR :
        {
            float elapsedTime = m_timer.getElapsedTime();

            if( elapsedTime != 0.0f )
            {
                frameMove( elapsedTime );
                render();
            }        
        }
#endif
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

int D3DApp::findPickedPos( int x, int y, D3DXVECTOR3 *vPicked )
{
    D3DVIEWPORT9 vp;
    m_pd3dDevice->GetViewport( &vp );

    D3DXVECTOR3 vRayDirNDC((float)x,(float)y,0.0f);
    D3DXVECTOR3 vRayDirB((float)x,(float)y,1.0f);
    D3DXVECTOR3 vNPt;
    D3DXVECTOR3 vNPtB;

    D3DXVec3Unproject(&vNPt,  &vRayDirNDC, &vp, &(m_camera.getMatProj()), &(m_camera.getMatView()), NULL);
    D3DXVec3Unproject(&vNPtB, &vRayDirB,   &vp, &(m_camera.getMatProj()), &(m_camera.getMatView()), NULL);

    vNPtB = vNPtB-vNPt;

    D3DXVECTOR3 *pEye = &vNPt;//&(m_camera.m_eye);//
    D3DXVECTOR3 *pRayDir = &vNPtB;

    return m_map.findPickedPos( pRayDir, pEye, vPicked );    
}

void D3DApp::frameMove( float elapsedTime )
{
    D3DXMATRIX worldViewProj = m_camera.getMatView() * m_camera.getMatProj();
    // move a character
    m_char.frameMove( elapsedTime, worldViewProj );
    // move a camera
    m_camera.move( m_char.getPosition() );
    
}

//
// render a scene
//
void D3DApp::render()
{
    // clear the backbuffer and the zbuffer
    m_pd3dDevice->Clear( 0,
                         NULL,
                         D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                         D3DCOLOR_XRGB( 0,0,0 ),
                         1.0f,
                         0 );
    // set FPS
    int fps = m_timer.getFPS();

    if( fps )
    {
        sprintf( strFrameRate,"FPS : %d", fps );
        Log::setFPSStr(strFrameRate);
    }

    // set camera
    m_camera.setCamera( m_pd3dDevice );

    if( SUCCEEDED(m_pd3dDevice->BeginScene()) )
    {
        // print log and fps
        Log::render( m_pd3dDevice );

        // render a map
        m_map.render( m_camera.getFrustum() );

        // render a char
        m_char.render();

        m_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}

//
// setup lights
//
void D3DApp::setupLights()
{
    // set up a white, directional light
    D3DXVECTOR3 vecDir;
    D3DLIGHT9 light;
    ZeroMemory( &light, sizeof( D3DLIGHT9 ) );
    light.Type = D3DLIGHT_DIRECTIONAL;
    D3DXCOLOR color( 1.0f, 1.0f, 1.0f, 1.0f );
    light.Ambient = color * 0.6f;
    light.Diffuse = color;
    light.Specular = color  * 0.6f;
    vecDir = D3DXVECTOR3( cosf( D3DX_PI/4),
                          -sinf( D3DX_PI/4),
                          0 );
    D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );
    light.Range = 10.0f;
    m_pd3dDevice->SetLight( 0, &light );
    m_pd3dDevice->LightEnable( 0, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE ); 
}